from math import sqrt

import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.misc import ascent
from skimage.filters import gaussian
from skimage.util import img_as_float32

from cyvlfeat.sift.dsift import dsift
from cyvlfeat.sift.sift import sift

img = ascent().astype(np.float32)


def test_dsift_slow_fast():
    # bin size in pixels
    bin_size = 4
    # bin size / keypoint scale
    magnif = 3
    scale = bin_size / magnif
    window_size = 5

    img_smooth = gaussian(img, sigma=sqrt(scale ** 2 - 0.25))
    _, d = dsift(img_smooth, size=bin_size, step=10,
                 window_size=window_size, float_descriptors=True)
    _, d_ = dsift(img_smooth, size=bin_size, step=10,
                  window_size=window_size, float_descriptors=True,
                  fast=True)
    err = np.std(d_ - d) / np.std(d)

    assert err < 0.1


@pytest.mark.parametrize('window_size', [5, 7.5, 10, 12.5, 20])
def test_dsift_sift(window_size):
    bin_size = 4
    magnif = 3
    scale = bin_size / magnif
    img_smooth = gaussian(img, sigma=sqrt(scale ** 2 - 0.25))
    f, d = dsift(img_smooth, size=bin_size,
                 step=10, window_size=window_size,
                 float_descriptors=True)
    num_keypoints = f.shape[0]
    f_ = np.column_stack([f, np.ones(shape=(num_keypoints,)) * scale, np.zeros(shape=(num_keypoints,))])
    f_, d_ = sift(img, magnification=magnif, frames=f_,
                  first_octave=-1, n_levels=5, compute_descriptor=True,
                  float_descriptors=True, window_size=window_size)
    err = np.std(d - d_) / np.std(d)

    assert err < 0.1


def test_dsift_non_float_descriptors():
    _, descriptors = dsift(img, float_descriptors=False)
    assert descriptors.dtype == np.uint8


def test_dsift_float_descriptors():
    _, descriptors = dsift(img, float_descriptors=True)
    assert descriptors.dtype == np.float32


def test_dsift_steps():
    # Step 3 in Y-Direction, 4 in X-Direction
    frames, descriptors = dsift(img, step=[3, 4])

    assert frames.shape[0] == 21168
    assert_allclose(frames[:3], [[4.5, 4.5],
                                 [4.5, 8.5],
                                 [4.5, 12.5]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, :10], [99, 0, 0, 0, 0, 0, 150, 24, 56, 0])


def test_dsift_windowsize():
    frames, descriptors = dsift(img, window_size=3)

    assert frames.shape[0] == 253009
    assert_allclose(frames[:3], [[4.5, 4.5],
                                 [4.5, 5.5],
                                 [4.5, 6.5]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, :10], [99, 0, 0, 0, 0, 0, 157, 24, 52, 0],
                    rtol=1e-3)


def test_dsift_norm():
    frames, descriptors = dsift(img, norm=True)

    assert frames.shape[-1] == 3
    assert frames.shape[0] == 253009
    print(frames)
    assert_allclose(frames[:3], [[4.5, 4.5, 0.2953],
                                 [4.5, 5.5, 0.2471],
                                 [4.5, 6.5, 0.2115]],
                    rtol=1e-3)
    assert_allclose(descriptors[0, :10], [99, 0, 0, 0, 0, 0, 150, 24, 56, 0],
                    rtol=1e-3)



import numpy as np
cimport numpy as np
cimport cython
# import cython
from libc.stdio cimport printf

# Import the header files
from cyvlfeat.cy_util cimport dtype_from_memoryview
from cyvlfeat._vl.host cimport VL_TYPE_FLOAT, VL_TYPE_DOUBLE
from cyvlfeat._vl.host cimport vl_size
from cyvlfeat._vl.vlad cimport vl_vlad_encode
from cyvlfeat._vl.vlad cimport VL_VLAD_FLAG_NORMALIZE_COMPONENTS
from cyvlfeat._vl.vlad cimport VL_VLAD_FLAG_SQUARE_ROOT
from cyvlfeat._vl.vlad cimport VL_VLAD_FLAG_UNNORMALIZED
from cyvlfeat._vl.vlad cimport VL_VLAD_FLAG_NORMALIZE_MASS

@cython.boundscheck(False)
cpdef cy_vlad(cython.floating[:, ::1] X,
            cython.floating[:, ::1] means,
            cython.floating[:, ::1] assignments,
            bint unnormalized,
            bint square_root,
            bint normalize_components,
            bint normalize_mass,
            bint verbose):
    dtype = dtype_from_memoryview(X)
    cdef:
        vl_size n_clusters = means.shape[0]
        vl_size n_dimensions = means.shape[1]
        vl_size n_data = X.shape[0]
        int flags = 0
        cython.floating[::1] enc = np.zeros(n_clusters * n_dimensions, dtype=dtype)

    if unnormalized:
        flags |= VL_VLAD_FLAG_UNNORMALIZED

    if normalize_components:
        flags |= VL_VLAD_FLAG_NORMALIZE_COMPONENTS

    if normalize_mass:
        flags |= VL_VLAD_FLAG_NORMALIZE_MASS

    if square_root:
        flags |= VL_VLAD_FLAG_SQUARE_ROOT

    if verbose:
        # check for 2 * n_clusters * n_dimensions in print spree
        print("vl_vlad: num data:             %d" % n_data)
        print("vl_vlad: num clusters:         %d" % n_clusters)
        print("vl_vlad: data dimension:       %d" % n_dimensions)
        print("vl_vlad: code dimension:       %d" % 2 * n_clusters * n_dimensions)
        print("vl_vlad: unnormalized:         %d" % unnormalized)
        print("vl_vlad: normalize mass:       %d" % normalize_mass)
        print("vl_vlad: normalize components: %d" % normalize_components)
        print("vl_vlad: square root:          %d" % square_root)

    vl_float_type = VL_TYPE_FLOAT if dtype == np.float32 else VL_TYPE_DOUBLE
    vl_vlad_encode(&enc[0],
                  vl_float_type,
                  &means[0, 0],
                  n_dimensions,
                  n_clusters,
                  &X[0, 0],
                  n_data,
                  &assignments[0, 0],
                  flags)

    return np.asarray(enc)



from cyvlfeat.vlad.vlad import vlad
import numpy as np
from numpy.testing import assert_allclose

def get_fake_data(dtype=np.float32):
    x = np.random.randn(256, 128)
    mu = np.random.randn(16, 128)
    assignments = np.random.uniform(size=(256, 16))
    assignments = assignments * (1 / assignments.sum(axis=1, keepdims=True))

    return x, mu, assignments

# the helper function is convert from vlfeat's vl_test_vlad
# note that MATLAB uses column-major order while numpy uses row-major order
# tranpose the input array when uses this function
def simple_vlad(x, mu, assign):
    enc = []
    for i in range(assign.shape[0]):
        enc.append(np.matmul(x, assign[i, :].transpose()) - assign[i, :].sum() * mu[:, i])
    enc = np.concatenate(enc)

    return enc

def test_basic():
    for dtype in [np.float32, np.float64]:
        x = np.array([[1], [2], [3]]).astype(dtype)
        mu = np.zeros(shape=(3, 1), dtype=dtype)
        assignments = np.eye(3, dtype=dtype)
        observed_enc = vlad(x, mu, assignments, unnormalized=True)
        expected_enc = np.array([1, 2, 3]).astype(dtype)
        assert_allclose(expected_enc, observed_enc)

def test_rand():
    for dtype in [np.float32, np.float64]:
        x, mu, assignments = get_fake_data(dtype)
        observed_enc = vlad(x, mu, assignments, unnormalized=True)
        expected_enc = simple_vlad(x.transpose(), mu.transpose(), assignments.transpose())
        assert_allclose(observed_enc, expected_enc)

def test_norm():
    for dtype in [np.float32, np.float64]:
        x, mu, assignments = get_fake_data(dtype=dtype)
        expected_enc = simple_vlad(x.transpose(), mu.transpose(), assignments.transpose())
        expected_enc = expected_enc / np.linalg.norm(expected_enc)
        observed_enc = vlad(x, mu, assignments)
        assert_allclose(observed_enc, expected_enc)

def test_sqrt():
    for dtype in [np.float32, np.float64]:
        x, mu, assignments = get_fake_data(dtype=dtype)
        expected_enc = simple_vlad(x.transpose(), mu.transpose(), assignments.transpose())
        expected_enc = np.sign(expected_enc) * np.sqrt(np.abs(expected_enc))
        expected_enc = expected_enc / np.linalg.norm(expected_enc)
        observed_enc = vlad(x, mu, assignments, square_root=True)
        assert_allclose(observed_enc, expected_enc)

def test_individual():
    for dtype in [np.float32, np.float64]:
        x, mu, assignments = get_fake_data(dtype=dtype)
        expected_enc = simple_vlad(x.transpose(), mu.transpose(), assignments.transpose())
        expected_enc = expected_enc.reshape((-1, x.shape[1]))
        expected_enc = expected_enc * (1 / np.sqrt(np.sum(expected_enc**2, axis=1, keepdims=True)))
        expected_enc = expected_enc.flatten()
        observed_enc = vlad(x, mu, assignments, unnormalized=True, normalize_components=True)
        assert_allclose(expected_enc, observed_enc)

def test_mass():
    for dtype in [np.float32, np.float64]:
        x, mu, assignments = get_fake_data(dtype=dtype)
        expected_enc = simple_vlad(x.transpose(), mu.transpose(), assignments.transpose())
        expected_enc = expected_enc.reshape((-1, x.shape[1]))
        expected_enc = expected_enc * np.transpose(1 / assignments.sum(axis=0, keepdims=True))
        expected_enc = expected_enc.flatten()
        observed_enc = vlad(x, mu, assignments, unnormalized=True, normalize_mass=True)
        assert_allclose(expected_enc, observed_enc)

def test_vlad_dimension():
    N = 1000
    K = 512
    D = 128

    x = np.random.uniform(size=(N, D)).astype(np.float32)
    means = np.random.uniform(size=(K, D)).astype(np.float32)
    assignments = np.random.uniform(size=(N, K)).astype(np.float32)
    enc = vlad(x, means, assignments)

    expected = K * D
    observed = len(enc)
    assert (expected == observed)



import numpy as np
import pytest
from scipy.misc import ascent, face

from cyvlfeat.quickshift.flatmap import flatmap
from cyvlfeat.quickshift.quickshift import quickshift

img = ascent().astype(np.float64)
img_rgb = face(gray=False).astype(np.float64)


def test_quickshift_rgb():
    result = quickshift(img_rgb,
                        kernel_size=4,
                        max_dist=24)
    assert result[0].dtype == np.float64
    assert result[1].dtype == np.float64
    assert result[2].dtype == np.float64
    assert result[0].shape == img_rgb.shape[:2]
    assert result[1].shape == img_rgb.shape[:2]
    assert result[2].shape == img_rgb.shape[:2]


@pytest.mark.parametrize('kernel_size', [2, 4, 8])
@pytest.mark.parametrize('max_dist', [None, 10, 20, 30])
@pytest.mark.parametrize('medoid', [True, False])
def test_quickshift(kernel_size, max_dist, medoid):
    result = quickshift(img,
                        kernel_size=kernel_size,
                        max_dist=max_dist,
                        medoid=medoid)
    assert result[0].dtype == np.float64
    assert result[1].dtype == np.float64
    assert result[0].shape == img.shape
    assert result[1].shape == img.shape
    if max_dist is not None:
        assert result[2].dtype == np.float64
        assert result[2].shape == img.shape


def test_flatmap():
    maps = quickshift(img, 2, 10)[0]
    labels, clusters = flatmap(maps)

    assert maps.dtype == np.float64
    assert maps.dtype == img.dtype
    assert maps.shape == img.shape
    assert labels.dtype == maps.dtype
    assert labels.shape == maps.shape
    assert clusters.dtype == maps.dtype
    assert clusters.size == maps.size



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size, vl_bool


cdef extern from "vl/quickshift.h":
    ctypedef double vl_qs_type

    ctypedef struct VlQS:
        vl_qs_type *image    # height x width x channels feature image
        int height           # height of the image
        int width            # width of the image
        int channels         # number of channels in the image
        vl_bool medoid
        vl_qs_type sigma
        vl_qs_type tau
        int *parents
        vl_qs_type *dists
        vl_qs_type *density

    VlQS *vl_quickshift_new(vl_qs_type *im, int height, int width, int channels)
    void vl_quickshift_delete(VlQS *q)
    void vl_quickshift_process (VlQS *q)

    vl_bool vl_quickshift_get_medoid(VlQS *q)
    int *vl_quickshift_get_parents(VlQS *q)
    vl_qs_type *vl_quickshift_get_dists(VlQS *q)
    vl_qs_type *vl_quickshift_get_density(VlQS *q)
    void vl_quickshift_set_max_dist(VlQS *f, vl_qs_type tau)
    void vl_quickshift_set_kernel_size(VlQS *f, vl_qs_type sigma)
    void vl_quickshift_set_medoid(VlQS *f, vl_bool medoid)




import numpy as np
from .cyhog import cy_hog


def hog(image, cell_size, variant='UoCTTI', n_orientations=9,
        directed_polar_field=False, undirected_polar_field=False,
        bilinear_interpolation=False, verbose=False, visualize=False,
        channels_first=False):
    r"""
    Computes the HOG features for ``image``
    and the specified ``cell_size``. ``image`` can be either greyscale or RGB
    and must be ``float32``.

    HOG decomposes an image into small squared cells, computes an histogram of
    oriented gradients in each cell, normalizes the result using a block-wise
    pattern, and return a descriptor for each cell.

    In this case the feature has 31 dimensions. HOG exists in many variants.
    VLFeat supports two: the UoCTTI variant (used by default) and the original
    Dalal-Triggs variant (with 2x2 square HOG blocks for normalization).
    The main difference is that the UoCTTI variant computes both directed and
    undirected gradients as well as a four dimensional texture-energy feature,
    but projects the result down to 31 dimensions.

    ``output`` is a ``float32`` array, its number
    of columns is approximately the number of columns of ``image`` divided
    by ``cell_size`` and the same for the number of rows. The third
    dimension spans the number of feature components.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] or [H, W, 3] `float32` `ndarray` (channel-last format)
        [1, H, W] or [3, H, W] `float32` `ndarray` (channel-first format),
        see also arg `channel_first`.
        A single channel, RGB or greyscale, `float32` numpy array (ndarray)
        representing the image to calculate descriptors for.
    cell_size : `int`
        The size of the cells that the image should be decomposed into. Assumed
        to be square and this only a single integer is required.
    variant : `str`, optional
        Choose a HOG variant from the set {'UoCTTI', 'DalalTriggs'}
    n_orientations : `int`, optional
        Choose a number of undirected orientations in the orientation
        histograms. The angle ``[0, pi)`` is divided in to ``n_orientations``
        equal parts.
    directed_polar_field : `bool`, optional
        By specifying this flag, the image is interpreted as samples
        from a 2D vector field. Angles are measure clockwise, the y-axis
        pointing downwards, starting from the x-axis (pointing to the right).
        The first channel should be the gradient magnitude and the second
        channel should be the directed (``[0, 2pi)``) orientations.
    undirected_polar_field : `bool`, optional
        By specifying this flag, the image is interpreted as samples
        from a 2D vector field as above, however, the angles will be forcibly
        wrapped into the range ``[0, pi)``.
    bilinear_interpolation : `bool`, optional
        This flags activates the use of bilinear interpolation to assign
        orientations to bins. This produces a smoother feature, but is
        only used for the 'DalalTriggs' variant.
    verbose : `bool`, optional
        If ``True``, be verbose.
    visualize : `bool`, optional
        Also return an image of the HOG. For each cell and orientation bin,
        the image contains a line segment that is centered at the cell center,
        is perpendicular to the midpoint of the range of angles spanned by the
        orientation bin, and has intensity proportional to the corresponding
        histogram value.
    channel_first : `bool`, optional
        VLfeat uses ``channel-first`` format by default, while ``channel-last``
        format is usually used in Python, e.g. scikit-image. Set this flag to
        `True` if your input is channel-first. cyvlfeat will convert input array
        from `channel-last` format to ``channel-first`` format internally if this
        flag is `False`.

    Returns
    -------
    output : `(RH, RW, D)` `float32` `ndarray`
        ``RH`` and ``RW`` is approximately ``H / cellsize`` and ``W / cellsize``. ``D`` is the number of feature dimensions.
    hog_image : 2D `uint8` `ndarray`, optional
        A visualisation of the HOG image. Only provided if `visualize` is `True`.
    """
    # vlfeat requires image in channel-first style
    if image.ndim == 3:
        if not channels_first:
            image = image.transpose([2, 0, 1])
    # Add a channels axis
    if image.ndim == 2:
        image = np.expand_dims(image, axis=0)

    # Validate image size
    if image.ndim != 3:
        raise ValueError('Only 2D arrays with a channels axis are supported')

    # Validate all the parameters
    if cell_size < 0:
        raise ValueError('cell_size must be > 0')
    if variant not in {'UoCTTI', 'DalalTriggs'}:
        raise ValueError("variant must be in set {'UoCTTI', 'DalalTriggs'}")
    if n_orientations < 0:
        raise ValueError('n_orientations must be > 0')
    if (directed_polar_field or undirected_polar_field) and image.shape[0] != 2:
        raise ValueError('Expected a polar field image of n_channels == 2')

    # Ensure types are correct before passing to Cython
    image = np.require(image, dtype=np.float32, requirements='C')

    # Shortcut for getting the correct enum value, since is_UoCTTI has
    # an enum value of 1 (True is 1 in Python)
    is_UoCTTI = variant == 'UoCTTI'

    result = cy_hog(image, cell_size, is_UoCTTI,
                    n_orientations, directed_polar_field,
                    undirected_polar_field, bilinear_interpolation,
                    True,  # Return channels as last axis
                    verbose, visualize)
    if visualize:
        viz_array = result[1]
        # properly rescale to 0-255 for visualization
        scale = 255 / (viz_array.max() - viz_array.min() + np.finfo(np.float32).eps)
        viz_array = scale * (viz_array - viz_array.min())
        viz_array = viz_array.astype(np.uint8)

        return result[0], viz_array
    else:
        return result



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.


import numpy as np
cimport numpy as np
cimport cython
from libc.stdio cimport printf
from libc.string cimport memcpy

# Import the header files
from cyvlfeat._vl.quickshift cimport *
from cyvlfeat._vl.host cimport *


@cython.boundscheck(False)
cpdef cy_quickshift(np.ndarray[double, ndim=3, mode='c'] image,
                    int kernel_size, int max_dist, bint compute_estimate,
                    bint medoid, bint verbose):
    cdef:
        int *parentsi
        double sigma = kernel_size
        double tau = max_dist
        int ndims = image.ndim
        VlQS *q

        int n1 = image.shape[1]
        int n2 = image.shape[2]

        # int num_channels = image.shape[2] if ndims == 3 else 1
        int num_channels = image.shape[0]

        # parent_c (same size as I) - Intermediate array
        np.ndarray[double, ndim=2, mode='c'] parents_c = np.empty(
            (n1*n2, 1), dtype=np.float64, order='C')

        # Create output arrays.
        # parents (same size as I)
        np.ndarray[double, ndim=2, mode='c'] parents = np.empty(
            (n1, n2), dtype=np.float64, order='C')

        # dists (same size as I)
        np.ndarray[double, ndim=2, mode='c'] dists = np.empty(
            (n1, n2), dtype=np.float64, order='C')

        # density (same size as I)
        np.ndarray[double, ndim=2, mode='c'] density = np.empty(
            (n1, n2), dtype=np.float64, order='C')


    if verbose:
        print("quickshift:   [N1, N2, K]:           = [%d, %d, %d]" % (n1, n2, num_channels))
        print("quickshift:   type:                  = %s" % 'medoid' if medoid else 'quick')
        print("quickshift:   kernel size:           = %g" % sigma)
        print("quickshift:   maximum gap:           = %g" % tau)

    q = vl_quickshift_new(&image[0, 0, 0], n1, n2, num_channels)
    vl_quickshift_set_kernel_size(q, sigma)
    vl_quickshift_set_max_dist(q, tau)
    vl_quickshift_set_medoid(q, medoid)

    vl_quickshift_process(q)

    parentsi = vl_quickshift_get_parents(q)

    # Copy results
    for i in range(n1*n2):
        parents_c[i][0] = parentsi[i] + 1

    parents = np.reshape(parents_c, (n1, n2))

    memcpy(&dists[0, 0], vl_quickshift_get_dists(q), sizeof(double) * n1 * n2)

    if compute_estimate:
        memcpy(&density[0, 0], vl_quickshift_get_density(q), sizeof(double) * n1 * n2)

    vl_quickshift_delete(q)

    if compute_estimate:
        return parents, dists, density
    else:
        return parents, dists


import os.path as osp

import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.misc import ascent
from scipy.spatial.distance import cdist

import cyvlfeat
from cyvlfeat.sift.sift import sift

# actual frames and descriptors are compute by sift from vlfeat binary
# command: sift -vv --frames --descriptors --first-octave -1 ascent.pgm
# where ascent.pgm is saved from `scipy.misc.ascent`

# by default `ascent()` return `int64`, but we need `np.float32`
img = ascent().astype(np.float32)
actual_frames = np.loadtxt(osp.join(osp.dirname(cyvlfeat.__file__), 'data', 'ascent.frame'), dtype=np.float32)
actual_frames = actual_frames[:, [1, 0, 2, 3]]
actual_descriptors = np.loadtxt(osp.join(osp.dirname(cyvlfeat.__file__), 'data', 'ascent.descr'), dtype=np.float32)


@pytest.mark.parametrize('compute_descriptor', [True, False])
def test_sift_shape(compute_descriptor):
    result = sift(img, compute_descriptor=compute_descriptor)
    if compute_descriptor:
        assert result[0].shape[1] == 4
        assert result[1].shape[1] == 128
        assert result[0].shape[0] == result[1].shape[0]
    else:
        assert result.shape[1] == 4


def test_sift_user_defined_frames():
    frames, descriptors = sift(img, frames=actual_frames, first_octave=-1,
                               compute_descriptor=True, float_descriptors=True,
                               norm_thresh=0, verbose=False)
    D2 = cdist(actual_frames, frames)
    index = D2.argmin(axis=1)
    descriptors = descriptors[index, :]
    t1 = np.mean(np.sqrt(np.sum((actual_descriptors - descriptors) ** 2, axis=1)))
    t2 = np.mean(np.sqrt(np.sum(actual_descriptors ** 2, axis=1)))
    err = t1 / t2

    assert err < 0.1


def test_sift_detector():
    f = sift(img, first_octave=-1, peak_thresh=0.01)
    frames = actual_frames.copy()
    # scale the components so that 1 pixel erro in x,y,z is equal to a
    # 10-th of angle
    scale = (20 / np.pi)
    frames[:, -1] = np.mod(frames[:, -1], np.pi * 2) * scale
    f[:, -1] = np.mod(f[:, -1], np.pi * 2) * scale
    D2 = cdist(frames, f)
    d2 = D2.min(axis=1)
    d2.sort()
    err = np.sqrt(d2)
    quant80 = round(0.8 * f.shape[0])

    # check for less than one pixel error at 80% quantile
    assert err[quant80] < 1


def test_sift_non_float_descriptors():
    frames, descriptors = sift(img, compute_descriptor=True)

    assert_allclose(frames[0], [5.28098, 342.099, 1.90079, 5.07138], rtol=1e-3)
    assert_allclose(descriptors[0, :5], [17, 0, 0, 11, 20])
    assert frames.shape[0] == 797


def test_sift_sort_user_defined_scales():
    frames = np.array([[5.28098, 342.099, 1.90079, 5.07138],
                       [6.90227, 242.036, 1.84193, 2.63054],
                       [22.1369, 273.583, 2.06255, 1.56508]])
    new_frames, descriptors = sift(img, frames=frames, compute_descriptor=True)

    assert new_frames.shape[0] == 3
    assert_allclose(new_frames[0], frames[1], rtol=1e-3)
    assert_allclose(new_frames[1], frames[0], rtol=1e-3)
    assert_allclose(new_frames[2], frames[2], rtol=1e-3)
    assert_allclose(descriptors[0, :10], [107, 171, 0, 0, 0, 0, 0, 11, 144, 171])
    assert_allclose(descriptors[1, :10], [17, 0, 0, 11, 20, 1, 16, 73, 95, 11])
    assert_allclose(descriptors[2, :10], [134, 15, 0, 0, 1, 1, 0, 0, 134, 73])


def test_sift_force_orientations():
    frames = np.array([[6.90227, 242.036, 1.84193, 2.63054],
                       [5.28098, 342.099, 1.90079, 5.07138],
                       [22.1369, 273.583, 2.06255, 1.56508]])
    new_frames, descriptors = sift(img, frames=frames, compute_descriptor=True,
                                   force_orientations=True)

    assert new_frames.shape[0] == 4
    assert_allclose(new_frames[0], [6.90227, 242.036, 1.84193, 2.63051], rtol=1e-3)
    assert_allclose(descriptors[0, :10], [107, 171, 0, 0, 0, 0, 0, 11, 144, 171])



import numpy as np
import pytest
from numpy.testing import assert_allclose
from scipy.misc import face
import cyvlfeat
from cyvlfeat.sift.phow import phow
import pytest


img_rgb = face(gray=False).astype(np.float32)
img_gray = face(gray=True).astype(np.float32)


@pytest.mark.parametrize('color', ['gray', 'rgb', 'hsv', 'opponent'])
def test_phow(color):
    if color == 'gray':
        frames, descriptors = phow(img_gray, color=color)
        assert descriptors.shape[1] == 128
    else:
        frames, descriptors = phow(img_rgb, color=color)
        assert descriptors.shape[1] == 128*3


# Copyright (C) 2013 David Novotny and Andera Vedaldi.
# All rights reserved.

# This file is part of the VLFeat library and is made available under
# the terms of the BSD license (see the COPYING file).

from cyvlfeat._vl.host cimport vl_size
from cyvlfeat._vl.host cimport vl_type

cdef extern from "vl/vlad.h":
    void vl_vlad_encode(void* enc,
                        vl_type dataType,
                        void* means,
                        vl_size dimension,
                        vl_size numClusters,
                        void* data,
                        vl_size numData,
                        void * assignments,
                        int flags)


    cdef int VL_VLAD_FLAG_NORMALIZE_COMPONENTS "VL_VLAD_FLAG_NORMALIZE_COMPONENTS"
    cdef int VL_VLAD_FLAG_SQUARE_ROOT "VL_VLAD_FLAG_SQUARE_ROOT"
    cdef int VL_VLAD_FLAG_UNNORMALIZED "VL_VLAD_FLAG_UNNORMALIZED"
    cdef int VL_VLAD_FLAG_NORMALIZE_MASS "VL_VLAD_FLAG_NORMALIZE_MASS"



import cyvlfeat
from cyvlfeat.hog import hog
import numpy as np
from numpy.testing import assert_allclose
from scipy.misc import face
from scipy.io import loadmat
import pytest
import os.path as osp

img_gray = face(gray=True).astype(np.float32)
img_rgb = face(gray=False).astype(np.float32)
# HOG features obtained from the MATLAB/Octave API
hog_features = loadmat(osp.join(osp.dirname(cyvlfeat.__file__), 'data', 'hog.mat'))


@pytest.mark.parametrize('cell_size', [8, 15, 24, 32])
@pytest.mark.parametrize('n_orientations', [4, 6, 9, 15])
@pytest.mark.parametrize('variant', ['DalalTriggs', 'UoCTTI'])
@pytest.mark.parametrize('visualize', [True, False])
def test_hog_shape(cell_size, n_orientations, variant, visualize):
    output = hog(img_rgb, cell_size=cell_size,
                 n_orientations=n_orientations,
                 variant=variant, visualize=visualize)
    h = int((img_rgb.shape[0] + cell_size / 2) // cell_size)
    w = int((img_rgb.shape[1] + cell_size / 2) // cell_size)
    if visualize:
        viz_h = 21 * h
        viz_w = 21 * w
    if variant == 'UoCTTI':
        c = 4 + 3 * n_orientations
    else:
        c = 4 * n_orientations
    if visualize:
        assert output[0].dtype == np.float32
        assert output[0].shape == (h, w, c)
        assert output[1].shape == (viz_h, viz_w)
    else:
        assert output.dtype == np.float32
        assert output.shape == (h, w, c)


def test_hog_cell_size_32_uoctti():
    des_gray = hog(img_gray, 32)
    des_rgb = hog(img_rgb, 32)

    assert_allclose(hog_features['hog_des_gray_uoctti'], des_gray, rtol=1e-3)
    assert_allclose(hog_features['hog_des_rgb_uoctti'], des_rgb, rtol=1e-3)


@pytest.mark.parametrize('bilinear_interpolation', [True, False])
def test_hog_cell_size_32_dalaltriggs(bilinear_interpolation):
    des_gray = hog(img_gray, 32, variant='DalalTriggs', bilinear_interpolation=bilinear_interpolation)
    des_rgb = hog(img_rgb, 32, variant='DalalTriggs', bilinear_interpolation=bilinear_interpolation)

    if bilinear_interpolation:
        assert_allclose(hog_features['hog_des_gray_dalaltriggs_bilinear'], des_gray, rtol=1e-4)
        assert_allclose(hog_features['hog_des_rgb_dalaltriggs_bilinear'], des_rgb, rtol=1e-4)
    else:
        assert_allclose(hog_features['hog_des_gray_dalaltriggs'], des_gray, rtol=1e-4)
        assert_allclose(hog_features['hog_des_rgb_dalaltriggs'], des_rgb, rtol=1e-4)

@pytest.mark.parametrize('variant', ['DalalTriggs', 'UoCTTI'])
def test_hog_channel_order(variant):
    img_rgb_chanel_first = img_rgb.transpose([2, 0, 1])
    des_rgb_channel_first, viz_rgb_channel_first = hog(img_rgb_chanel_first, 32, variant=variant,
                                                       channels_first=True, visualize=True)
    des_rgb, viz_rgb = hog(img_rgb, 32, variant=variant, channels_first=False, visualize=True)

    assert des_rgb.shape == des_rgb_channel_first.shape
    assert viz_rgb.shape == viz_rgb_channel_first.shape
    assert_allclose(des_rgb, des_rgb_channel_first)
    assert_allclose(viz_rgb, viz_rgb_channel_first)



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from cyvlfeat._vl.host cimport vl_bool


cdef extern from "vl/dsift.h":
    ctypedef struct VlDsiftKeypoint:
        double x        #< x coordinate */
        double y        #< y coordinate */
        double s        #< scale */
        double norm     #< SIFT descriptor norm */

    # @brief Dense SIFT descriptor geometry */
    ctypedef struct VlDsiftDescriptorGeometry:
        int numBinT     #< number of orientation bins */
        int numBinX     #< number of bins along X */
        int numBinY     #< number of bins along Y */
        int binSizeX    #< size of bins along X */
        int binSizeY    #< size of bins along Y */

    ctypedef struct VlDsiftFilter:
        int imWidth     #< @internal @brief image width */
        int imHeight    #< @internal @brief image height */

        int stepX       #< frame sampling step X */
        int stepY       #< frame sampling step Y */

        int boundMinX   #< frame bounding box min X */
        int boundMinY   #< frame bounding box min Y */
        int boundMaxX   #< frame bounding box max X */
        int boundMaxY   #< frame bounding box max Y */

        # descriptor parameters */
        VlDsiftDescriptorGeometry geom

        int useFlatWindow  #< flag: whether to approximate the Gaussian window with a flat one */
        double windowSize  #< size of the Gaussian window */

        int numFrames               #< number of sampled frames */
        int descrSize               #< size of a descriptor */
        VlDsiftKeypoint *frames     #< frame buffer */
        float *descrs               #< descriptor buffer */

        int numBinAlloc     #< buffer allocated: descriptor size */
        int numFrameAlloc   #< buffer allocated: number of frames  */
        int numGradAlloc    #< buffer allocated: number of orientations */

        float **grads       #< gradient buffer */
        float *convTmp1     #< temporary buffer */
        float *convTmp2     #< temporary buffer */

    VlDsiftFilter *vl_dsift_new(int width, int height)
    void vl_dsift_delete(VlDsiftFilter *self)
    void vl_dsift_process(VlDsiftFilter *self, float*im)

    void vl_dsift_set_steps(VlDsiftFilter *self,
                            int stepX,
                            int stepY)
    void vl_dsift_set_bounds(VlDsiftFilter *self,
                             int minX,
                             int minY,
                             int maxX,
                             int maxY)
    void vl_dsift_set_geometry(VlDsiftFilter *self,
                               VlDsiftDescriptorGeometry *geom)
    void vl_dsift_set_flat_window(VlDsiftFilter *self,
                                  vl_bool useFlatWindow)
    void vl_dsift_set_window_size(VlDsiftFilter *self, double windowSize)

    float* vl_dsift_get_descriptors(VlDsiftFilter *self)
    int  vl_dsift_get_descriptor_size(VlDsiftFilter *self)
    int vl_dsift_get_keypoint_num(VlDsiftFilter *self)
    VlDsiftKeypoint* vl_dsift_get_keypoints(VlDsiftFilter *self)
    void vl_dsift_get_bounds(VlDsiftFilter *self,
                             int *minX,
                             int *minY,
                             int *maxX,
                             int *maxY)
    void vl_dsift_get_steps(VlDsiftFilter*self,
                            int*stepX,
                            int*stepY)
    VlDsiftDescriptorGeometry*vl_dsift_get_geometry(VlDsiftFilter*self)
    double vl_dsift_get_window_size(VlDsiftFilter *self)



import os
import platform
import site

from setuptools import setup, find_packages, Extension

SYS_PLATFORM = platform.system().lower()
IS_LINUX = 'linux' in SYS_PLATFORM
IS_OSX = 'darwin' == SYS_PLATFORM
IS_WIN = 'windows' == SYS_PLATFORM

# Get Numpy include path without importing it
NUMPY_INC_PATHS = [os.path.join(r, 'numpy', 'core', 'include')
                   for r in site.getsitepackages() if
                   os.path.isdir(os.path.join(r, 'numpy', 'core', 'include'))]
if len(NUMPY_INC_PATHS) == 0:
    try:
        import numpy as np
    except ImportError:
        raise ValueError("Could not find numpy include dir and numpy not installed before build - "
                         "cannot proceed with compilation of cython modules.")
    else:
        # just ask numpy for it's include dir
        NUMPY_INC_PATHS = [np.get_include()]

elif len(NUMPY_INC_PATHS) > 1:
    print("Found {} numpy include dirs: "
          "{}".format(len(NUMPY_INC_PATHS), ', '.join(NUMPY_INC_PATHS)))
    print("Taking first (highest precedence on path): {}".format(
        NUMPY_INC_PATHS[0]))
NUMPY_INC_PATH = NUMPY_INC_PATHS[0]


# ---- C/C++ EXTENSIONS ---- #
# Stolen (and modified) from the Cython documentation:
#     http://cython.readthedocs.io/en/latest/src/reference/compilation.html
def no_cythonize(extensions, **_ignore):
    import os.path as op
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in ('.pyx', '.py'):
                if extension.language == 'c++':
                    ext = '.cpp'
                else:
                    ext = '.c'
                sfile = path + ext
                if not op.exists(sfile):
                    raise ValueError('Cannot find pre-compiled source file '
                                     '({}) - please install Cython'.format(sfile))
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


def build_extension_from_pyx(pyx_path, extra_sources_paths=None):
    # If we are building from the conda folder,
    # then we know we can manually copy some files around
    # because we have control of the setup. If you are
    # building this manually or pip installing, you must satisfy
    # that the vlfeat vl folder is on the PATH (for the headers)
    # and that the vl.dll file is visible to the build system
    # as well.
    include_dirs = [NUMPY_INC_PATH]
    library_dirs = []
    if IS_WIN:
        include_dirs.append(os.environ['LIBRARY_INC'])
        library_dirs.append(os.environ['LIBRARY_BIN'])

    if extra_sources_paths is None:
        extra_sources_paths = []
    extra_sources_paths.insert(0, pyx_path)
    ext = Extension(name=pyx_path[:-4].replace('/', '.'),
                    sources=extra_sources_paths,
                    include_dirs=include_dirs,
                    library_dirs=library_dirs,
                    libraries=['vl'],
                    language='c')
    if IS_LINUX or IS_OSX:
        ext.extra_compile_args.extend([
            '-Wno-unused-function',
            '-Wno-implicit-function-declaration'
    ])
    if IS_OSX:
        ext.extra_link_args.append('-headerpad_max_install_names')
    return ext


try:
    from Cython.Build import cythonize
except ImportError:
    import warnings

    cythonize = no_cythonize
    warnings.warn('Unable to import Cython - attempting to build using the '
                  'pre-compiled C++ files.')

cython_modules = [
    build_extension_from_pyx('cyvlfeat/fisher/cyfisher.pyx'),
    build_extension_from_pyx('cyvlfeat/generic/generic.pyx'),
    build_extension_from_pyx('cyvlfeat/gmm/cygmm.pyx'),
    build_extension_from_pyx('cyvlfeat/hog/cyhog.pyx'),
    build_extension_from_pyx('cyvlfeat/kmeans/cykmeans.pyx'),
    build_extension_from_pyx('cyvlfeat/quickshift/cyquickshift.pyx'),
    build_extension_from_pyx('cyvlfeat/sift/cysift.pyx'),
    build_extension_from_pyx('cyvlfeat/vlad/cyvlad.pyx'),
]
cython_exts = cythonize(cython_modules, quiet=True, language_level="3")


def get_version_and_cmdclass(package_name):
    import os
    from importlib.util import module_from_spec, spec_from_file_location
    spec = spec_from_file_location('version',
                                   os.path.join(package_name, '_version.py'))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass('cyvlfeat')
setup(
    name='cyvlfeat',
    version=version,
    cmdclass=cmdclass,
    description='Cython wrapper of the VLFeat toolkit',
    url='https://github.com/menpo/cyvlfeat',
    author='Patrick Snape',
    author_email='patricksnape@gmail.com',
    ext_modules=cython_exts,
    package_data={'cyvlfeat': ['data/*.mat', 'data/ascent.descr', 'data/ascent.frame']},
    packages=find_packages()
)



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
cimport cython

# Import the header files
from cyvlfeat._vl.hog cimport *
from cyvlfeat._vl.host cimport VL_FALSE, vl_size
from cyvlfeat.cy_util cimport set_python_vl_printf


@cython.boundscheck(False)
cpdef cy_hog(float[:, :, ::1] data, vl_size cell_size, int variant,
             vl_size n_orientations, bint directed_polar_field,
             bint undirected_polar_field, bint bilinear_interpolation,
             bint return_channels_last_axis, bint verbose, bint visualize):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        # Python images are not transposed
        # RGB image is channel first, aka. (C, H, W)
        VlHog* hog = vl_hog_new(<VlHogVariant>variant,
                                n_orientations, VL_FALSE)
        vl_size height = data.shape[1]
        vl_size width = data.shape[2]
        vl_size n_channels = data.shape[0]
        vl_size out_height = 0, out_width = 0, out_n_channels = 0
        float[:, :, ::1] out_array
        float[:, ::1] viz_array
        vl_size viz_glyph_size

    vl_hog_set_use_bilinear_orientation_assignments(hog,
                                                    bilinear_interpolation)

    if directed_polar_field or undirected_polar_field:
        vl_hog_put_polar_field(hog,
                               &data[0, 0, 0],  # Magnitude
                               &data[1, 0, 0],  # Angle
                               directed_polar_field,
                               width, height, cell_size)
    else:  # Assume we have an image
        vl_hog_put_image(hog, &data[0, 0, 0], width, height, n_channels,
                         cell_size)

    out_height = vl_hog_get_height(hog)
    out_width = vl_hog_get_width(hog)
    out_n_channels = vl_hog_get_dimension(hog)

    if verbose:
        print('vl_hog: image: [%lld x %lld x %lld]' % (height, width, n_channels))
        print('vl_hog: descriptor: [%lld x %lld x %lld]' % out_height, out_width, out_n_channels)
        print('vl_hog: number of orientations: %lld' % (n_orientations))
        print('vl_hog: bilinear orientation assignments: %s' %
              'yes' if bilinear_interpolation else 'no')
        print('vl_hog: variant: %s' %
              'DalalTriggs' if variant == VlHogVariantDalalTriggs else 'UOCTTI')
        print('vl_hog: input type: %s' %
              'DirectedPolarField' if directed_polar_field
              else ('UndirectedPolarField' if undirected_polar_field else 'Image'))

    # according to https://www.vlfeat.org/api/hog.html
    # hog features array is (out_n_channels, out_height, out_width) with "C" order
    out_array = np.zeros((out_n_channels, out_height, out_width),
                         dtype=np.float32, order='C')

    vl_hog_extract(hog, &out_array[0][0][0])

    if visualize:
        viz_glyph_size = vl_hog_get_glyph_size(hog)
        viz_width = viz_glyph_size * vl_hog_get_width(hog)
        viz_height = viz_glyph_size * vl_hog_get_height(hog)
        viz_array = np.zeros((viz_glyph_size * out_height,
                            viz_glyph_size * out_width),
                            dtype=np.float32, order='C')
        vl_hog_render(hog, &viz_array[0][0], &out_array[0][0][0],
                      out_width, out_height)
        if verbose:
            print("vl_hog: glyph size: %lld" % viz_glyph_size)
            print("vl_hog: glyph image: [%lld x %lld]" %
                  viz_glyph_size * out_height,
                  viz_glyph_size * out_width)

    vl_hog_delete(hog)
    # we prefer (out_height, out_width, out_n_channels) in numpy
    if return_channels_last_axis:
        out_array = np.transpose(out_array, [1, 2, 0]).copy()
    if visualize:
        return np.asarray(out_array), np.asarray(viz_array)
    else:
        return np.asarray(out_array)



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size


cdef extern from "vl/sift.h":
    ctypedef float vl_sift_pix

    ctypedef struct VlSiftKeypoint:
        int o           # < o coordinate(octave).
        int ix          # < Integer unnormalized x coordinate.
        int iy          # < Integer unnormalized y coordinate.
        int c_is "is"   # < Integer s coordinate.
        float x         # < x coordinate.
        float y         # < y coordinate.
        float s         # < s coordinate.
        float sigma     # < scale.

    ctypedef struct VlSiftFilt:
        double sigman               # < nominal image smoothing.
        double sigma0               # < smoothing of pyramid base.
        double sigmak               # < k-smoothing
        double dsigma0              # < delta-smoothing.
        int width                   # < image width.
        int height                  # < image height.
        int O                       # < number of octaves.
        int S                       # < number of levels per octave.
        int o_min                   # < minimum octave index.
        int s_min                   # < minimum level index.
        int s_max                   # < maximum level index.
        int o_cur                   # < current octave.
        vl_sift_pix *temp           # < temporary pixel buffer.
        vl_sift_pix *octave         # < current GSS data.
        vl_sift_pix *dog            # < current DoG data.
        int octave_width            # < current octave width.
        int octave_height           # < current octave height.
        vl_sift_pix *gaussFilter    # < current Gaussian filter
        double gaussFilterSigma     # < current Gaussian filter std
        vl_size gaussFilterWidth    # < current Gaussian filter width
        VlSiftKeypoint*keys         # < detected keypoints.
        int nkeys                   # < number of detected keypoints.
        int keys_res                # < size of the keys buffer.
        double peak_thresh          # < peak threshold.
        double edge_thresh          # < edge threshold.
        double norm_thresh          # < norm threshold.
        double magnif               # < magnification factor.
        double windowSize           # < size of Gaussian window(in spatial bins)
        vl_sift_pix *grad           # < GSS gradient data.
        int grad_o                  # < GSS gradient data octave.

    VlSiftFilt* vl_sift_new(int width, int height, int noctaves, int nlevels,
                            int o_min)
    void vl_sift_delete(VlSiftFilt *f)
    int vl_sift_process_first_octave(VlSiftFilt *f, vl_sift_pix *im)
    int vl_sift_process_next_octave(VlSiftFilt *f)
    void vl_sift_detect(VlSiftFilt *f)
    int vl_sift_calc_keypoint_orientations(VlSiftFilt *f, double *angles,
                                           VlSiftKeypoint *k)
    void vl_sift_calc_keypoint_descriptor(VlSiftFilt *f, vl_sift_pix *descr,
                                          VlSiftKeypoint *k, double angle)
    void vl_sift_keypoint_init(VlSiftFilt *f, VlSiftKeypoint *k, double x,
                               double y, double sigma)
    int vl_sift_get_octave_index(VlSiftFilt *f)
    int vl_sift_get_noctaves(VlSiftFilt *f)
    int vl_sift_get_octave_first(VlSiftFilt *f)
    int vl_sift_get_nlevels(VlSiftFilt *f)
    int vl_sift_get_nkeypoints(VlSiftFilt *f)
    double vl_sift_get_peak_thresh(VlSiftFilt *f)
    double vl_sift_get_edge_thresh(VlSiftFilt *f)
    double vl_sift_get_norm_thresh(VlSiftFilt *f)
    double vl_sift_get_magnif(VlSiftFilt *f)
    double vl_sift_get_window_size(VlSiftFilt *f)
    VlSiftKeypoint *vl_sift_get_keypoints(VlSiftFilt *f)
    void vl_sift_set_peak_thresh(VlSiftFilt *f, double t)
    void vl_sift_set_edge_thresh(VlSiftFilt *f, double t)
    void vl_sift_set_norm_thresh(VlSiftFilt *f, double t)
    void vl_sift_set_magnif(VlSiftFilt *f, double m)
    void vl_sift_set_window_size(VlSiftFilt *f, double m)
